from .wiktionary_normseg import CASE_NORMSEG_MAP

MOOD_MAP = {"cond": "-isi", "impv": "(!)", "potn": "-ne", "opt": "-os"}

TENSE_MAP = {"past": "-i", "present": None}

PERS_MAP = {
    "pl1": "-mme", "pl2": "-tte", "pl3": "-vat", "sg1": "-n", "sg2": "-t", "sg3": None
}

CASE_NAME_MAP = {
    "abe": "abessive",
    "abl": "ablative",
    "ade": "adessive",
    "all": "allative",
    "com": "comitative",
    "ela": "elative",
    "ess": "essive",
    "gen": "genitive",
    "ill": "illative",
    "ine": "inessive",
    "ins": "instructive",
    "nom": "nominative",
    "par": "partitive",
    "tra": "translative",
    "lat": None,  # "-s"?
    "acc": None,  # -n / -ut ?
}

INF_MAP = {"e": "-e", "ma": "-ma", "minen": "-minen"}

CASE_MAP = {
    k: CASE_NORMSEG_MAP[p] for k, p in CASE_NAME_MAP.items() if p in CASE_NORMSEG_MAP
}

NUM_MAP = {"pl": "-t", "sg": None}

POSS_MAP = {"sg1": "-ni", "sg2": "-si", "pl1": "-mme", "pl2": "-nne", "3": "-nsa"}



from tempfile import TemporaryDirectory
from finntk.utils import ResourceMan, urlretrieve
import os

# If we don't ensure this, we might end up with a relative path for LEMMA_FILENAME
os.makedirs(os.path.expanduser("~/.conceptnet5"), exist_ok=True)
from conceptnet5.language.lemmatize import LEMMA_FILENAME  # noqa


class ConceptNetWiktionaryResMan(ResourceMan):
    RESOURCE_NAME = "conceptnet_wiktionary"

    URL = "https://archive.org/download/wiktionary.db/wiktionary.db.gz"

    def __init__(self):
        self._cached_is_bootstrapped = None
        super().__init__()

    def _is_bootstrapped(self):
        if self._cached_is_bootstrapped is not None:
            return self._cached_is_bootstrapped
        is_boostrapped = os.path.exists(LEMMA_FILENAME)
        self._cached_is_bootstrapped = is_boostrapped
        return is_boostrapped

    def _bootstrap(self, _res=None):
        from plumbum.cmd import gunzip

        tempdir = TemporaryDirectory()
        en_wiktionary_gz = urlretrieve(
            self.URL, filename=os.path.join(tempdir.name, "enwiktionary.gz")
        )
        try:
            gunzip(en_wiktionary_gz, LEMMA_FILENAME)
        finally:
            os.remove(en_wiktionary_gz)
            tempdir.cleanup()
        self._cached_is_bootstrapped = True


conceptnet_wiktionary = ConceptNetWiktionaryResMan()



# OMorFi's Python library can output UD but there's no way to do the reverse
# map: UD to Omor


PASSTHROUGHS = {"CASE", "NUMTYPE", "PRONTYPE", "ADPTYPE", "FOREIGN"}
PASSTHROUGHS_KEY_MAP = {
    "Clitic": "CLIT",
    "Degree": "CMP",
    "Person": "PERS",
    "Person[psor]": "POSS",
    # XXX: Except perhaps this should sometimes be LEX
    "Derivation": "DRV",
}
NUM_KEY_MAP = {"Number": "PERS", "Number[psor]": "POSS"}

NUM_VAL_MAP = {"Sing": "SG", "Plur": "PL"}

TENSE_MAP = {"Pres": "PRESENT", "Past": "PAST"}

MOOD_MAP = {
    "Ind": "INDV",
    "Cnd": "COND",
    "Imp": "IMPV",
    # Passthrough
}

VOICE_MAP = {"Pass": "PSS", "Act": "ACT"}

PART_FORM_MAP = {
    "Pres": "VA", "Past": "NUT", "Agent": "MA", "Agt": "MA", "Neg": "MATON"
}

INF_FORM_MAP = {"1": "A", "2": "E", "3": "MA", "4": "MINEN", "5": "MAISILLA"}



import wordfreq
import numpy as np
from sklearn.preprocessing import normalize
from sklearn.decomposition import TruncatedSVD
from numpy.ma import size


def power_mean_inner(p, mat, axis=0):
    mat_cmplx = mat.astype(np.complex128)
    result_cmplx = np.power(
        (np.power(mat_cmplx, p)).sum(axis=axis) / size(mat_cmplx, axis=axis), 1 / p
    )
    return result_cmplx.real.astype(np.float_)


def power_mean(p, mat):
    if p == float("inf"):
        return mat.max(axis=0)
    elif p == float("-inf"):
        return mat.min(axis=0)
    elif p == 1:
        return mat.mean(axis=0)
    else:
        return power_mean_inner(p, mat, axis=0)


def bow_to_mat(space, stream):
    refs = []
    vecs = []
    for ref in stream:
        vec = space.get_vec(ref)
        if vec is None:
            continue
        refs.append(ref)
        vecs.append(vec)
    if not refs:
        return refs, None
    return refs, np.stack(vecs)


CATP_3 = (float("-inf"), 1, float("inf"))
CATP_4 = (float("-inf"), 1, 3, float("inf"))


def catp_mean(mat, ps=CATP_3):
    """
    From *Concatenated p-mean Word Embeddings as Universal Cross-Lingual
    Sentence Representations*
    https://arxiv.org/pdf/1803.01400.pdf
    """
    p_means = []
    for p in ps:
        p_means.append(power_mean(p, mat))
    return np.hstack(p_means)


def compute_pc(X):
    svd = TruncatedSVD(n_components=1, n_iter=7, random_state=0)
    svd.fit(X)
    return svd.components_


def pre_sif_mean_inner(mat, freqs, a):
    """
    From *A Simple but Tough-to-Beat Baseline for Sentence Embeddings*
    https://openreview.net/forum?id=SyK00v5xx
    https://github.com/PrincetonML/SIF
    """
    # 1. Normalize
    mat = normalize(mat)
    # 2. Reweight
    rows, cols = mat.shape
    for coli, freq in enumerate(freqs):
        mat[coli, :] = (a / (a + freq)) * mat[coli, :]
    return mat.mean(axis=0)


def get_wf(maybe_pair):
    if isinstance(maybe_pair, tuple):
        return maybe_pair[0]
    return maybe_pair


def pre_sif_mean(mat, refs, lang):
    return pre_sif_mean_inner(
        mat, (wordfreq.word_frequency(get_wf(ref), lang) for ref in refs), 1e-3
    )


pre_sif_mean.needs_words = True


def mk_sif_mean(pc):

    def sif_mean(mat, refs, lang):
        emb = pre_sif_mean(mat, refs, lang)
        return emb - emb.dot(pc) * pc

    sif_mean.needs_words = True
    return sif_mean


def unnormalized_mean(mat):
    return mat.mean(axis=0)


def normalized_mean(mat):
    normalized_mat = normalize(mat)
    return normalized_mat.mean(axis=0)


def apply_vec(func, vec_getter, stream, lang=None, **extra):
    refs, mat = bow_to_mat(vec_getter, stream)
    if mat is None:
        return
    if hasattr(func, "needs_words"):
        return func(mat, refs, lang, **extra)
    else:
        return func(mat, **extra)


def cosine_sim(u, v):
    from scipy.spatial.distance import cosine

    return 1 - cosine(u, v)


def get(space, ref):
    try:
        return space.get_vector(ref)
    except KeyError:
        pass



import ahocorasick
from finntk.data.wordnet import ALL_ABBRVS

_abbrv_auto = None


def get_abbrv_auto():
    global _abbrv_auto
    if _abbrv_auto is not None:
        return _abbrv_auto
    _abbrv_auto = ahocorasick.Automaton()
    for abbrv in ALL_ABBRVS:
        _abbrv_auto.add_word("_{}_".format(abbrv), abbrv)
    _abbrv_auto.make_automaton()
    return _abbrv_auto


def has_abbrv(lemma):
    """
    Given a FinnWordNet formatted lemma, e.g. saada_tehdä_jtak return whether
    it contains a placeholder abbreviation.
    """
    abbrv_auto = get_abbrv_auto()
    it = abbrv_auto.iter("_{}_".format(lemma))
    return len(list(it))



def main():
    from finntk.emb.fasttext import vecs as fasttext_vecs
    from finntk.emb.autoextend import vecs as autoextend_vecs
    from finntk.emb.numberbatch import vecs as numberbatch_vecs
    from finntk.wordnet.reader import fiwn_resman

    for res_man in [fasttext_vecs, autoextend_vecs, numberbatch_vecs, fiwn_resman]:
        for resource in res_man.resource_names():
            res_man.bootstrap(resource)


if __name__ == "__main__":
    main()



"""
Function to get ahold of an OMorFi instance.
"""

_omorfi = None


def get_omorfi():
    """
    Gets an Omorfi instance with everything possible enabled. Reuses the
    existing instance if already called once.
    """
    from omorfi.omorfi import Omorfi

    global _omorfi
    if _omorfi is None:
        _omorfi = Omorfi(use_describe=True)
        _omorfi.load_from_dir(
            analyse=True,
            generate=True,
            accept=True,
            tokenise=True,
            lemmatise=True,
            hyphenate=True,
            segment=True,
            labelsegment=True,
            guesser=True,
            udpipe=True,
            describe=True,
        )
    return _omorfi



"""
Implementation of Lesk++, which *may* be broken.
"""
from finntk.emb.autoextend import vecs as fiwn_vecs, mk_lemma_vec
import numpy as np
from .lesk_emb import MultilingualLesk


class LeskPP:

    def __init__(self, multispace, aggf, wn_filter=False, expand=False):
        self.multispace = multispace
        self.aggf = aggf
        self.wn_filter = wn_filter
        self.expand = expand
        self.ml_lesk = MultilingualLesk(multispace, aggf, wn_filter, expand)

    def mk_defn_vec(self, item):
        return self.ml_lesk.mk_defn_vec(item)

    def mk_ctx_vec(self, sent_lemmas, exclude_idx=None):
        fiwn_space = fiwn_vecs.get_vecs()

        words = []
        vecs = []
        for idx, (lemma_str, lemmas) in enumerate(sent_lemmas):
            if exclude_idx is not None and idx == exclude_idx:
                continue
            vec_appended = False
            if len(lemmas) == 1:
                try:
                    vecs.append(mk_lemma_vec(lemmas[0]))
                except KeyError:
                    pass
                else:
                    vec_appended = True
            if not vec_appended:
                try:
                    vecs.append(fiwn_space.get_vector(lemma_str))
                except KeyError:
                    pass
                else:
                    vec_appended = True
            if vec_appended:
                words.append(lemma_str)

        if not vecs:
            return
        mat = np.stack(vecs)
        if hasattr(self.aggf, "needs_words"):
            return self.aggf(mat, words, "fi")
        else:
            return self.aggf(mat)



import numpy as np
from typing import Tuple

from .base import RefType
from .fasttext import multispace as fasttext_ms
from .numberbatch import multispace as numberbatch_ms
from .word2vec import space as word2vec_fi
from .base import MultilingualVectorSpace, MonoVectorSpaceAdapter


def none_hstack(*dim_vecs):
    if all(vec is None for dim, vec in dim_vecs):
        return None
    return np.hstack(
        [vec if vec is not None else np.zeros(dim) for dim, vec in dim_vecs]
    )


class FastTextNumberbatchMultiSpace(MultilingualVectorSpace):
    takes = RefType.BOTH
    dim = fasttext_ms.dim + numberbatch_ms.dim

    def get_vec(self, lang: str, ref: Tuple[str, str]):
        wf, lemma = ref
        return none_hstack(
            (fasttext_ms.dim, fasttext_ms.get_vec(lang, wf)),
            (numberbatch_ms.dim, numberbatch_ms.get_vec(lang, lemma)),
        )


ft_nb_multispace = FastTextNumberbatchMultiSpace()


fasttext_fi = MonoVectorSpaceAdapter(fasttext_ms, "fi")
numberbatch_fi = MonoVectorSpaceAdapter(numberbatch_ms, "fi")


class FastTextNumberbatchWord2VecFiSpace(MultilingualVectorSpace):
    takes = RefType.BOTH
    dim = fasttext_fi.dim + numberbatch_fi.dim + word2vec_fi.dim

    def get_vec(self, ref: Tuple[str, str]):
        wf, lemma = ref
        return none_hstack(
            (fasttext_fi.dim, fasttext_fi.get_vec(wf)),
            (numberbatch_fi.dim, numberbatch_fi.get_vec(lemma)),
            (word2vec_fi.dim, word2vec_fi.get_vec(wf)),
        )


ft_nb_w2v_space = FastTextNumberbatchWord2VecFiSpace()



STD_ABBRVS = {
    "jstak": "jostakin",
    "jssak": "jossakin",
    "jnak": "jonakin",
    "jtak": "jotakin",
    "jllak": "jollakin",
    "jltak": "joltakin",
    "jksta": "jostakusta",
    "jllek": "jollekin",
    "jkssa": "jossakussa",
    "jkta": "jotakuta",
    "jklla": "jollakulla",
    "jklta": "joltakulta",
    "jhk": "johonkin",
    "jklle": "jollekulle",
    "jksik": "joksikin",
    "jkksi": "joksikuksi",
    "jkna": "jonakuna",
    "jkhun": "johonkuhun",
    "jku": "joku",  # (“somebody”).
    "jk": "jokin",  # (“something”).
    "jkn": "jonkun",  # (“of somebody, somebody's”).
    "jnk": "jonkin",
}

OTHER_ABBRVS = {
    "jhkin": "johonkin",
    "jhkun": "johonkuhun",
    "jksk": "joksikin",
    "jkskin": "joksikin",
    "jlkn": "jollakin",
    "jllk": "jollekin",
    "jltk": "joltakin",
    "jnnk": "jonnekin",  # new
    "jsatk": "jostakin",
    "jssk": "jossakin",
    "jstk": "jostakin",
    "jtk": "jotakin",
}

ALL_ABBRVS = {**STD_ABBRVS, **OTHER_ABBRVS}

PRON_CASE = {
    "jostakin": "ela",
    "jossakin": "ine",
    "jonakin": "ess",
    "jotakin": "par",
    "jollakin": "ade",
    "joltakin": "abl",
    "jostakusta": "ela",
    "jollekin": "all",
    "jossakussa": "ine",
    "jotakuta": "par",
    "jollakulla": "ade",
    "joltakulta": "abl",
    "johonkin": "ill",
    "jollekulle": "all",
    "joksikin": "tra",
    "joksikuksi": "tra",
    "jonakuna": "ess",
    "johonkuhun": "ill",
    "joku": "nom",
    "jokin": "nom",
    "jonkun": "gen",
    "jonkin": "gen",
    "jonnekin": "ill",
}

PRON_LEMMAS = {
    "jostakin": "jokin",
    "jossakin": "jokin",
    "jonakin": "jokin",
    "jotakin": "jokin",
    "jollakin": "jokin",
    "joltakin": "jokin",
    "jostakusta": "joku",
    "jollekin": "jokin",
    "jossakussa": "joku",
    "jotakuta": "joku",
    "jollakulla": "joku",
    "joltakulta": "joku",
    "johonkin": "jokin",
    "jollekulle": "joku",
    "joksikin": "joku",
    "joksikuksi": "joku",
    "jonakuna": "joku",
    "johonkuhun": "joku",
    "joku": "joku",
    "jokin": "jokin",
    "jonkun": "joku",
    "jonkin": "jokin",
    "jonnekin": "jokin",
}



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy
import logging

import torch
from torch.utils.data import TensorDataset, DataLoader, SequentialSampler

from pytorch_pretrained_bert.tokenization import BertTokenizer, BasicTokenizer
from pytorch_pretrained_bert.modeling import BertModel

logging.basicConfig(format = '%(asctime)s - %(levelname)s - %(name)s -   %(message)s', 
                    datefmt = '%m/%d/%Y %H:%M:%S',
                    level = logging.INFO)
logger = logging.getLogger(__name__)


class InputExample(object):

    def __init__(self, unique_id, text_a, text_b):
        self.unique_id = unique_id
        self.text_a = text_a
        self.text_b = text_b


class InputFeatures(object):
    """A single set of features of data."""

    def __init__(self, unique_id, tokens, input_ids, input_mask, input_type_ids):
        self.unique_id = unique_id
        self.tokens = tokens
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.input_type_ids = input_type_ids


def convert_examples_to_features(examples, seq_length, tokenizer):
    """Loads a data file into a list of `InputBatch`s."""

    features = []
    for (ex_index, example) in enumerate(examples):
        tokens_a = tokenizer.tokenize(example.text_a)

        tokens_b = None
        if example.text_b:
            tokens_b = tokenizer.tokenize(example.text_b)

        if tokens_b:
            # Modifies `tokens_a` and `tokens_b` in place so that the total
            # length is less than the specified length.
            # Account for [CLS], [SEP], [SEP] with "- 3"
            _truncate_seq_pair(tokens_a, tokens_b, seq_length - 3)
        else:
            # Account for [CLS] and [SEP] with "- 2"
            if len(tokens_a) > seq_length - 2:
                tokens_a = tokens_a[0:(seq_length - 2)]

        # The convention in BERT is:
        # (a) For sequence pairs:
        #  tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
        #  type_ids: 0   0  0    0    0     0       0 0    1  1  1  1   1 1
        # (b) For single sequences:
        #  tokens:   [CLS] the dog is hairy . [SEP]
        #  type_ids: 0   0   0   0  0     0 0
        #
        # Where "type_ids" are used to indicate whether this is the first
        # sequence or the second sequence. The embedding vectors for `type=0` and
        # `type=1` were learned during pre-training and are added to the wordpiece
        # embedding vector (and position vector). This is not *strictly* necessary
        # since the [SEP] token unambigiously separates the sequences, but it makes
        # it easier for the model to learn the concept of sequences.
        #
        # For classification tasks, the first vector (corresponding to [CLS]) is
        # used as as the "sentence vector". Note that this only makes sense because
        # the entire model is fine-tuned.
        tokens = []
        input_type_ids = []
        tokens.append("[CLS]")
        input_type_ids.append(0)
        for token in tokens_a:
            tokens.append(token)
            input_type_ids.append(0)
        tokens.append("[SEP]")
        input_type_ids.append(0)

        if tokens_b:
            for token in tokens_b:
                tokens.append(token)
                input_type_ids.append(1)
            tokens.append("[SEP]")
            input_type_ids.append(1)

        input_ids = tokenizer.convert_tokens_to_ids(tokens)

        # The mask has 1 for real tokens and 0 for padding tokens. Only real
        # tokens are attended to.
        input_mask = [1] * len(input_ids)

        # Zero-pad up to the sequence length.
        while len(input_ids) < seq_length:
            input_ids.append(0)
            input_mask.append(0)
            input_type_ids.append(0)

        assert len(input_ids) == seq_length
        assert len(input_mask) == seq_length
        assert len(input_type_ids) == seq_length

        if ex_index < 5:
            logger.info("*** Example ***")
            logger.info("unique_id: %s" % (example.unique_id))
            logger.info("tokens: %s" % " ".join([str(x) for x in tokens]))
            logger.info("input_ids: %s" % " ".join([str(x) for x in input_ids]))
            logger.info("input_mask: %s" % " ".join([str(x) for x in input_mask]))
            logger.info(
                "input_type_ids: %s" % " ".join([str(x) for x in input_type_ids]))

        features.append(
            InputFeatures(
                unique_id=example.unique_id,
                tokens=tokens,
                input_ids=input_ids,
                input_mask=input_mask,
                input_type_ids=input_type_ids))
    return features


def _truncate_seq_pair(tokens_a, tokens_b, max_length):
    """Truncates a sequence pair in place to the maximum length."""

    # This is a simple heuristic which will always truncate the longer sequence
    # one token at a time. This makes more sense than truncating an equal percent
    # of tokens from each, since if one sequence is very short then each token
    # that's truncated likely contains more information than a longer sequence.
    while True:
        total_length = len(tokens_a) + len(tokens_b)
        if total_length <= max_length:
            break
        if len(tokens_a) > len(tokens_b):
            tokens_a.pop()
        else:
            tokens_b.pop()


def read_sents(sents):
    """Read a list of `InputExample`s from an input file."""
    examples = []
    unique_id = 0
    for sent in sents:
        text_a = sent
        text_b = None
        examples.append(
            InputExample(unique_id=unique_id, text_a=text_a, text_b=text_b))
        unique_id += 1
    return examples


def get_bert(model_path, use_cuda=True, do_lower_case=False):
    device = torch.device("cuda" if torch.cuda.is_available() and use_cuda else "cpu")
    n_gpu = torch.cuda.device_count()
    logger.info("device: {} n_gpu: {} ".format(device, n_gpu))

    tokenizer = BertTokenizer.from_pretrained(model_path, do_lower_case=do_lower_case)

    model = BertModel.from_pretrained(model_path)
    model.to(device)
    model.eval()
    model.device = device

    return tokenizer, model


class PassthroughBasicTokenizer(BasicTokenizer):
    def tokenize(self, tokens):
        """Tokenizes a piece of text."""
        output_tokens = []
        for token in tokens:
            token = self._clean_text(token)
            if self.do_lower_case:
                token = token.lower()
                token = self._run_strip_accents(token)
            output_tokens.append(token)

        return output_tokens


def embed_sentences(tok_model, sents, output_layer=-1, batch_size=64, max_seq_length=128, tokenize=True):
    tokenizer, model = tok_model
    if not tokenize:
        tokenizer.basic_tokenizer = PassthroughBasicTokenizer(tokenizer.basic_tokenizer.do_lower_case)

    examples = read_sents(sents)
    features = convert_examples_to_features(
        examples=examples, seq_length=max_seq_length, tokenizer=tokenizer)

    all_input_ids = torch.tensor([f.input_ids for f in features], dtype=torch.long)
    all_input_mask = torch.tensor([f.input_mask for f in features], dtype=torch.long)
    all_example_index = torch.arange(all_input_ids.size(0), dtype=torch.long)

    eval_data = TensorDataset(all_input_ids, all_input_mask, all_example_index)
    eval_sampler = SequentialSampler(eval_data)
    eval_dataloader = DataLoader(eval_data, sampler=eval_sampler, batch_size=batch_size)

    res = []

    for input_ids, input_mask, example_indices in eval_dataloader:
        input_ids = input_ids.to(model.device)
        input_mask = input_mask.to(model.device)

        all_encoder_layers, _ = model(input_ids, token_type_ids=None, attention_mask=input_mask)
        # float64 because sklearn's ball_tree uses it later on
        all_encoder_layers_np = numpy.empty((len(all_encoder_layers),) + all_encoder_layers[0].shape, numpy.float64)
        for layer_idx, layer in enumerate(all_encoder_layers):
            all_encoder_layers_np[layer_idx] = layer.detach().cpu().numpy()

        for b, example_index in enumerate(example_indices):
            feature = features[example_index.item()]
            sent_output = all_encoder_layers_np[:, b, :, :]
            tokens = feature.tokens
            res.append((sent_output, tokens))
    return res



from sklearn.neighbors import NearestNeighbors
import numpy as np
from os import makedirs
from os.path import join as pjoin, exists
import pickle
import logging


logger = logging.getLogger(__name__)


def normalize(vec):
    vec_norm = np.linalg.norm(vec)
    if not np.nonzero(vec_norm):
        return None
    return vec / vec_norm


class WordExpertBase:

    def __init__(self, algorithm="ball_tree"):
        self.ys = []
        self.clf = None
        self.algorithm = algorithm

    def add_word(self, ctx_vec, sense_key):
        self.ys.append(sense_key)

    def fit(self):
        clf = NearestNeighbors(n_neighbors=1, algorithm=self.algorithm)
        clf.fit(self.xs, self.ys)
        self.clf = clf
        del self.xs

    def predict(self, ctx_vec):
        indices = self.clf.kneighbors([ctx_vec], return_distance=False)
        index = indices[0, 0]
        return self.ys[index]


class OverfilledFixedWordExpert(Exception):
    pass


class FixedWordExpert(WordExpertBase):
    """
    This uses np.float64 because it's what ball_tree uses
    """

    def __init__(self, size, algorithm="ball_tree"):
        self.xs = None
        self.x_idx = 0
        self.size = size
        super().__init__(algorithm=algorithm)

    def fit(self):
        if self.xs is not None:
            self.xs.resize(self.x_idx, self.xs.shape[1])
        else:
            self.xs = [[]]
        super().fit()

    def add_word(self, ctx_vec, sense_key):
        if self.x_idx > self.size:
            raise OverfilledFixedWordExpert(
                f"Trying to add a word to a full FixedWordExpert with capacity: {self.size}"
            )
        normed = normalize(ctx_vec)
        if normed is None:
            # Actually don't add it
            return
        if self.xs is None:
            self.xs = np.ndarray((self.size, ctx_vec.shape[0]), dtype=np.float64)
        self.xs[self.x_idx] = normed
        self.x_idx += 1
        super().add_word(ctx_vec, sense_key)


class VarWordExpert(WordExpertBase):

    def __init__(self, algorithm="ball_tree"):
        self.xs = []
        super().__init__(algorithm=algorithm)

    def add_word(self, ctx_vec, sense_key):
        normed = normalize(ctx_vec)
        if normed is None:
            # Actually don't add it
            return
        self.xs.append(normed)
        super().add_word(ctx_vec, sense_key)


class ExpertExists(Exception):
    pass


class WordExpertManager:

    def __init__(self, path, mode="r", allow_overwrites=False):
        self.path = path
        self.allow_overwrites = allow_overwrites
        if mode == "w":
            makedirs(path, exist_ok=allow_overwrites)

    def _expert_path(self, word):
        return pjoin(self.path, word)

    def dump_expert(self, word, expert):
        expert_path = self._expert_path(word)
        if not self.allow_overwrites and exists(expert_path):
            raise ExpertExists(f"Not overwriting existing word expert: {expert_path}")
        logger.debug(f"Dumping word expert {word}; {expert} to {expert_path}")
        with open(expert_path, "wb") as outf:
            pickle.dump(expert, outf, protocol=4)
        logger.debug("Dumped")

    def load_expert(self, word):
        expert_path = self._expert_path(word)
        logger.debug(f"Loading {word} from {expert_path}")
        if not exists(expert_path):
            logger.debug("(doesn't exist)")
            return
        with open(expert_path, "rb") as inf:
            return pickle.load(inf)



"""
Mappings from terms used on English Wiktionary to normalised morphemes
"""

TEMPLATE_NORMSEG_MAP = {
    "comparative of": "-mpi",
    "superlative of": "-in",
    "agent noun of": "-ja",
    "plural of": "-t",
}

CASE_NORMSEG_MAP = {
    "nominative": None,
    "genitive": "-n",
    "partitive": "-ta",
    "inessive": "-ssa",
    "elative": "-sta",
    "illative": "-seen",  # XXX: -an? -Vn?
    "adessive": "-lla",
    "ablative": "-lta",
    "allative": "-lle",
    "essive": "-na",
    "translative": "-ksi",
    "instructive": "-in",
    "abessive": "-tta",
    "comitative": "-ine",
}

PL_CASES = {"instructive", "comitative"}

PL_NORMSEG_MAP = {"singular": None, "plural": "-t"}

NOUN_FORM_OF_FIELDS_MAP = {"case": CASE_NORMSEG_MAP, "pl": PL_NORMSEG_MAP}

VERB_FORM_OF_FIELDS_MAP = {
    "mood": {
        "indicative": None,
        "conditional": "-isi",
        "imperative": "(!)",
        "potential": "-ne",
        "optative": "-os",
    },
    "tense": {"past": "-i", "present": None, "present connegative": "(ei)"},
    ("pr", "pl"): {  # pr and pl need to be considered together
        ("first-person", "plural"): "-mme",
        ("second-person", "plural"): "-tte",
        ("third-person", "plural"): "-vat",
        ("impersonal", "plural"): "-taan",
        ("first-person", "singular"): "-n",
        ("second-person", "singular"): "-t",
        ("third-person", "singular"): None,
        ("impersonal", "singular"): "-taan",
    },
    "suffix": None,  # pass through
}

PARTICIPLES_MAP = {
    "pres": "-va",
    "pres_pass": "-ttava",
    "past": "-nut",
    "past_pass": "-ttu",
    "agnt": "-ma",
    "nega": ("-ma", "-ton"),
}

PARTICIPLES_NORM = {"pres_pasv": "pres_pass", "past_pasv": "past_pass"}

# XXX: Is this identical to CASE_NAME_MAP in omorfi_normseg?
FI_INFINITIVE_OF_ABBRVS = {
    "nom": "nominative",
    "gen": "genitive",
    "par": "partitive",
    "acc": "accusative",
    "ine": "inessive",
    "ela": "elative",
    "ill": "illative",
    "ade": "adessive",
    "abl": "ablative",
    "all": "allative",
    "ess": "essive",
    "tra": "translative",
    "ins": "instructive",
    "abe": "abessive",
    "com": "comitative",
}

FI_INFINITIVES = {
    "1l": "-ksi",
    "2a": "-e",
    "2p": "-tae",
    "3a": "-ma",
    "3p": "-tama",
    "4": "-minen",
    "5": ("-ma", "-isi"),
}

FI_INFINITIVE_DEFAULT_CASES = {
    "1l": "illative",
    "2a": "inessive",
    "2p": "inessive",
    "3a": None,
    "3p": "instructive",
    "4": "nominative",
    "5": "adessive",
}



import zipfile
import logging
from finntk.utils import ResourceMan, urlretrieve
import os

logger = logging.getLogger(__name__)


class ElmoWordVecs(ResourceMan):
    RESOURCE_NAME = "elmo-fi"

    URL = "http://vectors.nlpl.eu/repository/11/149.zip"

    def __init__(self):
        super().__init__()
        self._model = None

    def _get_res_filename(self, res):
        return res

    def _bootstrap(self, res):
        logger.info("Downloading Elmo word vectors")
        zipped_tmp_fn = urlretrieve(self.URL)
        try:
            tmp_zip = zipfile.ZipFile(zipped_tmp_fn)
            tmp_zip.extractall(self.get_res(""))
        finally:
            os.remove(zipped_tmp_fn)

    def get(self):
        from finntk.vendor.elmo import get_elmo

        if self._model is None:
            logger.info("Loading ELMo")
            use_cuda = not os.environ.get("DISABLE_CUDA")
            self._model = get_elmo(self.get_res(""), use_cuda=use_cuda)
            logger.info("Loaded ELMo")
        return self._model


vecs = ElmoWordVecs()



from nltk.util import binary_search_file
import zipfile
import logging
from functools import total_ordering
from finntk.utils import ResourceMan, urlretrieve
from finntk.wordnet.reader import fiwn
from gensim.models import KeyedVectors
from shutil import copyfileobj
import os

logger = logging.getLogger(__name__)


class AutoExtendNumberBatchFiWNWordVecs(ResourceMan):
    RESOURCE_NAME = "autoextend-fi"

    URL = (
        "https://github.com/frankier/AutoExtend/releases/download/"
        + "fiwn-conceptnet/fiwn-conceptnet.zip"
    )

    def __init__(self):
        super().__init__()
        self._resources["vecs"] = "fiwn.numberbatch.binvec"
        self._resources["synsets"] = "synsets.txt"
        self._vecs = None

    def _bootstrap(self, res):
        from gensim.test.utils import get_tmpfile

        logger.info("Downloading FiWN ConceptNet word vectors")
        zipped_tmp_fn = urlretrieve(self.URL)
        try:
            tmp_zip = zipfile.ZipFile(zipped_tmp_fn)
            tmp_fn = get_tmpfile("fiwn-conceptnet.txt")
            try:
                copyfileobj(tmp_zip.open("outputVectors.txt"), open(tmp_fn, "wb"))
                logger.info("Converting FiWN ConceptNet word vectors")
                fi = KeyedVectors.load_word2vec_format(tmp_fn)
                fi.save(self._get_res_path("vecs"))
            finally:
                os.remove(tmp_fn)
            copyfileobj(
                tmp_zip.open("synsets.txt"), open(self._get_res_path("synsets"), "wb")
            )
        finally:
            os.remove(zipped_tmp_fn)

    def get_vecs(self):
        if self._vecs is None:
            vec_path = self.get_res("vecs")
            logger.info("Loading word vectors")
            self._vecs = KeyedVectors.load(vec_path, mmap="r")
            logger.info("Loaded word vectors")
        return self._vecs


vecs = AutoExtendNumberBatchFiWNWordVecs()


@total_ordering
class AsKey(object):

    def __init__(self, obj, key_func):
        self.obj = obj
        self.obj_key = key_func(obj)
        self.key_func = key_func

    def __lt__(self, other):
        return self.obj_key < self.key_func(other)

    def __eq__(self, other):
        return self.obj_key == self.key_func(other)

    def __add__(self, other):
        return AsKey(self.obj + other.encode("utf-8"), self.key_func)

    def __len__(self):
        return len(self.obj)


POS_ORDER = b"nvar"


def synset_map_key(line):
    key = line.split(b" ", 1)[0]
    _, off, pos = key.strip().rsplit(b"-", 2)
    return (POS_ORDER.index(pos), off)


class SynsetMapper:

    def __init__(self, res_man):
        self.res_man = res_man
        self._map_file = None

    @property
    def map_file(self):
        if self._map_file is None:
            synsets_fn = self.res_man.get_res("synsets")
            self._map_file = open(synsets_fn, "rb")

            def discard_line(self):
                self.readline()

            self._map_file.discard_line = discard_line
        return self._map_file

    def __call__(self, synset_id):
        full_synset_id = "wn-fi-2.0-" + synset_id
        line = binary_search_file(
            self.map_file, AsKey(full_synset_id.encode("utf-8"), synset_map_key)
        )
        if line is None:
            return
        bits = line.split(b" ", 1)
        return bits[1].rstrip().decode("utf-8") or None


synset_map = SynsetMapper(vecs)


def get_lemma_id(lemma):
    synset = lemma.synset()
    return "{}-wn-fi-2.0-{}".format(lemma.name().lower(), fiwn.ss2of(synset))


def mk_lemma_vec(lemma):
    fiwn_space = vecs.get_vecs()
    return fiwn_space[get_lemma_id(lemma)]


def mk_lemmas_mat(lemmas):
    fiwn_space = vecs.get_vecs()
    return fiwn_space[[get_lemma_id(lemma) for lemma in lemmas]]


def mk_synset_vec(synset):
    fiwn_space = vecs.get_vecs()
    synset_id = synset_map(fiwn.ss2of(synset))
    if synset_id is None:
        return
    return fiwn_space[synset_id]



from nltk.corpus.reader.wordnet import Synset, WordNetError
from typing import Optional


def pre_id_to_post(pre_id):
    """
    Convert a synset id of the format n88888888 to 88888888-n
    """
    return "{}-{}".format(pre_id[1:], pre_id[0])


def post_id_to_pre(post_id, conv_s=False):
    """
    Convert a synset id of the format 88888888-n to n88888888
    """
    pos = post_id[-1]
    if conv_s and pos == "s":
        pos = "a"
    return "{}{}".format(pos, post_id[:-2])


def ss2pre(ss):
    pos = ss.pos()
    if pos == "s":
        pos = "a"
    return ("{}{:08d}".format(pos, ss.offset()))


def pre2ss(wn, pre):
    return wn.synset_from_pos_and_offset(pre[0], int(pre[1:]))


def fi2en_post(post_id):
    from .reader import get_en_fi_maps

    fi2en, en2fi = get_en_fi_maps()
    return pre_id_to_post(fi2en[post_id_to_pre(post_id, conv_s=True)])


def en2fi_post(post_id):
    from .reader import get_en_fi_maps

    fi2en, en2fi = get_en_fi_maps()
    return pre_id_to_post(en2fi[post_id_to_pre(post_id, conv_s=True)])


def maybe_fi2en_ss(ss: Synset) -> Optional[Synset]:
    from .reader import get_en_fi_maps
    from .utils import ss2pre, pre2ss
    from nltk.corpus import wordnet

    fi2en, _en2fi = get_en_fi_maps()
    pre_fi = ss2pre(ss)
    pre_en = fi2en.get(pre_fi)
    if pre_en is None:
        return None
    try:
        return pre2ss(wordnet, pre_en)
    except WordNetError:
        return None
